/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include "Camera.h"

Camera::Camera(const double& fov,  const double& aspect,  const double& near,  const double& far): fov(fov), aspect(aspect), near(near), far(far), m_pos(gk::Vector(7.0f,3.0f,3.0f)), m_target(gk::Vector(0.0f,0.0f,0.0f)), m_up(gk::Vector(0.0f,1.0f,0.0f)), speed(0.1), sensivity(0.2), angleX(0), angleY(0){
  Camera::model= gk::RotateY(0.f);
  Camera::projection= gk::Perspective(fov, aspect, near, far);   
}

Camera::Camera(const double& fov,  const double& aspect,  const double& near,  const double& far, const gk::Vector& position,const gk::Vector& look,const gk::Vector& up, double speed, double sensivity, double angleX, double angleY):fov(fov), aspect(aspect), near(near), far(far), m_pos(position), m_target(look), m_up(up), speed(speed), sensivity(sensivity), angleX(angleX), angleY(angleY){
  Camera::model= gk::RotateY(0.f); 
  Camera::projection= gk::Perspective(fov, aspect, near, far);   
  Update();
}

Camera::Camera(const Camera& c):fov(c.fov), aspect(c.aspect), near(c.near), far(c.far),m_pos(c.m_pos), m_target(c.m_target), m_up(c.m_up), m_forward(c.m_forward), speed(c.speed), sensivity(c.sensivity), angleX(c.angleX), angleY(c.angleY), model(c.model), projection(c.projection), viewBox(c.viewBox), volume(c.volume){}

Camera::~Camera(){}



void Camera::advance(int s){
  m_pos += (m_forward * speed*s); 
  m_target = m_pos + m_forward;
}

void Camera::back(int s){
  m_pos -= (m_forward * speed*s);
  m_target = m_pos + m_forward;
}

void Camera::left(int s){
  m_pos += m_left * speed * s;
  m_target = m_pos + m_forward;
}

void Camera::right(int s){
  m_pos -= m_left * speed * s;
  m_target = m_pos + m_forward;
}

void Camera::up(int s){
  m_pos += m_up * speed * s;
  m_target = m_pos + m_forward;
}

void Camera::down(int s){
  m_pos -= m_up * speed * s;
  m_target = m_pos + m_forward;
}

void Camera::Update(){

  if (angleY > 89)
    angleY = 89;
  else if (angleY < -89)
    angleY = -89;
  double r_temp = cos(angleY*M_PI/180);
  m_forward.y = sin(angleY*M_PI/180);
  m_forward.x = r_temp*cos(angleX*M_PI/180);
  m_forward.z = r_temp*sin(angleX*M_PI/180);

  m_left = Cross(m_up,m_forward);
  m_left = Normalize(m_left);


  m_target = m_pos + m_forward;
  

  gk::Matrix4x4 mvp =  getMVP().matrix();
  float frustum[6][4];
  float t;
/* Extract the numbers for the RIGHT plane */
   frustum[0][0] = mvp.m[3][0] - mvp.m[0][0];
   frustum[0][1] = mvp.m[3][1] - mvp.m[0][1];
   frustum[0][2] = mvp.m[3][2] - mvp.m[0][2];
   frustum[0][3] = mvp.m[3][3] - mvp.m[0][3];

   /* Normalize the result */
   t = sqrt( frustum[0][0] * frustum[0][0] + frustum[0][1] * frustum[0][1] + frustum[0][2] * frustum[0][2] );
   frustum[0][0] /= t;
   frustum[0][1] /= t;
   frustum[0][2] /= t;
   frustum[0][3] /= t;

   /* Extract the numbers for the LEFT plane */
   frustum[1][0] = mvp.m[3][0] + mvp.m[0][0];
   frustum[1][1] = mvp.m[3][1] + mvp.m[0][1];
   frustum[1][2] = mvp.m[3][2] + mvp.m[0][2];
   frustum[1][3] = mvp.m[3][3] + mvp.m[0][3];

   /* Normalize the result */
   t = sqrt( frustum[1][0] * frustum[1][0] + frustum[1][1] * frustum[1][1] + frustum[1][2] * frustum[1][2] );
   frustum[1][0] /= t;
   frustum[1][1] /= t;
   frustum[1][2] /= t;
   frustum[1][3] /= t;

   /* Extract the BOTTOM plane */
   frustum[2][0] = mvp.m[3][0] + mvp.m[1][0];
   frustum[2][1] = mvp.m[3][1] + mvp.m[1][1];
   frustum[2][2] = mvp.m[3][2] + mvp.m[1][2];
   frustum[2][3] = mvp.m[3][3] + mvp.m[1][3];

   /* Normalize the result */
   t = sqrt( frustum[2][0] * frustum[2][0] + frustum[2][1] * frustum[2][1] + frustum[2][2] * frustum[2][2] );
   frustum[2][0] /= t;
   frustum[2][1] /= t;
   frustum[2][2] /= t;
   frustum[2][3] /= t;

   /* Extract the TOP plane */
   frustum[3][0] = mvp.m[3][0] - mvp.m[1][0];
   frustum[3][1] = mvp.m[3][1] - mvp.m[1][1];
   frustum[3][2] = mvp.m[3][2] - mvp.m[1][2];
   frustum[3][3] = mvp.m[3][3] - mvp.m[1][3];

   /* Normalize the result */
   t = sqrt( frustum[3][0] * frustum[3][0] + frustum[3][1] * frustum[3][1] + frustum[3][2] * frustum[3][2] );
   frustum[3][0] /= t;
   frustum[3][1] /= t;
   frustum[3][2] /= t;
   frustum[3][3] /= t;

   /* Extract the FAR plane */
   frustum[4][0] = mvp.m[3][0] - mvp.m[2][0];
   frustum[4][1] = mvp.m[3][1] - mvp.m[2][1];
   frustum[4][2] = mvp.m[3][2] - mvp.m[2][2];
   frustum[4][3] = mvp.m[3][3] - mvp.m[2][3];

   /* Normalize the result */
   t = sqrt( frustum[4][0] * frustum[4][0] + frustum[4][1] * frustum[4][1] + frustum[4][2] * frustum[4][2] );
   frustum[4][0] /= t;
   frustum[4][1] /= t;
   frustum[4][2] /= t;
   frustum[4][3] /= t;

   /* Extract the NEAR plane */
   frustum[5][0] = mvp.m[3][0] + mvp.m[2][0];
   frustum[5][1] = mvp.m[3][1] + mvp.m[2][1];
   frustum[5][2] = mvp.m[3][2] + mvp.m[2][2];
   frustum[5][3] = mvp.m[3][3] + mvp.m[2][3];

   /* Normalize the result */
   t = sqrt( frustum[5][0] * frustum[5][0] + frustum[5][1] * frustum[5][1] + frustum[5][2] * frustum[5][2] );
   frustum[5][0] /= t;
   frustum[5][1] /= t;
   frustum[5][2] /= t;
   frustum[5][3] /= t;

  volume[RIGHT].setCoefficients(
				frustum[0][0],
				frustum[0][1],
				frustum[0][2],
				frustum[0][3]);
  volume[LEFT].setCoefficients(
				frustum[1][0],
				frustum[1][1],
				frustum[1][2],
				frustum[1][3]);

  volume[BOTTOM].setCoefficients(
				frustum[2][0],
				frustum[2][1],
				frustum[2][2],
				frustum[2][3]);

  volume[TOP].setCoefficients(
				frustum[3][0],
				frustum[3][1],
				frustum[3][2],
				frustum[3][3]);

  volume[FRONT].setCoefficients(
				frustum[4][0],
				frustum[4][1],
				frustum[4][2],
				frustum[4][3]);

  volume[BACK].setCoefficients(
				frustum[5][0],
				frustum[5][1],
				frustum[5][2],
				frustum[5][3]);


}






int Camera::visible(const std::array<gk::Vector,8>& b){
  int result = 1, out,in;
  for(int i=0; i < 6; i++) {
    out=0;
    in=0;
    for (int k = 0; k < 8; k++) {
      if (volume[i].distance(b[k]) < 0)
	out++;
      else
	in++;
    }
    //if all corners are out
    if (!in)
      return 0;
    // if some corners are out and others are in
    else if (out)
      result = 2;
  }
  return result;

}

void Camera::rotate(int x, int y){
  angleY -= y*sensivity;
  angleX += x*sensivity;
  Update();
}

std::ostream & operator <<(std::ostream & os, const Camera& c){
  for(int i = 0; i<6; i++)
    os<<c.volume[i];
  return os;
}


