/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#ifndef CAMERA2_H
#define CAMERA2_H
#include <cmath>
#include <array>
#include "Transform.h"
#include "Geometry.h"
#include "Plane.h"
#define ToRadian(x) (float)(((x) * M_PI / 180.0f))
#define ToDegree(x) (float)(((x) * 180.0f / M_PI))
#define ANG2RAD 3.14159265358979323846/180.0

#define MARGIN 10
#define TOP 0
#define FRONT 1
#define BACK 2
#define LEFT 3
#define RIGHT 4
#define BOTTOM 5


class Camera
{

 private:
  double fov, aspect, near, far;

  gk::Vector m_pos;
  gk::Vector m_target;
  gk::Vector m_up;
  gk::Vector m_forward;
  gk::Vector m_left; 
  double speed;
  double sensivity;
  double angleX;
  double angleY;
  gk::Transform model, projection;
  std::array<gk::Vector,8> viewBox;
  std::array<Plane,6> volume;
  void Update();
 protected:
 public:
  Camera(const double&,  const double&,  const double&,  const double&);
  Camera( const double&,  const double&,  const double&,  const double&, const gk::Vector&,const gk::Vector&,const gk::Vector&, double=0.1, double=0.2,double=0.0, double=0.0);
  Camera(const Camera&);
  ~Camera();
  
  //  const gk::Transform& view();
  void rotate(int, int);
  void advance(int=1);
  void back(int=1);
  void right(int=1);
  void left(int=1);
  void up(int=1);
  void down(int=1);

  int visible(const std::array<gk::Vector,8>&);

  inline const gk::Vector& getPosition() const {
    return m_pos;
  }

  inline const gk::Vector& getTarget() const {
    return m_target;
  }

  inline const gk::Vector& getUp() const {
    return m_up;
  }

  inline const gk::Vector& getLeft() const {
    return m_left;
  }
  inline const gk::Transform getModel() const {
    return model;
  }
  inline gk::Transform getView() const {
    return gk::LookAt(gk::Point(m_pos),gk::Point(m_target),m_up);
  }
  inline const gk::Transform& getProjection() const {
    return projection;
  }
  inline const gk::Transform getMV() const {
    return  getView() * model;
  }
  inline const gk::Transform getMVP() const {
    return projection * getView() * model;
  }

  inline const gk::Transform getVP() const {
    return getView()*projection ;
  }

  inline const gk::Transform getPV() const {
    return projection * getView() ;
  }

  inline const std::array<gk::Vector,8>& getViewBox() const {
    return viewBox;
  }
  
  inline const std::array<Plane,6>& getVolume() const {
    return volume;
  }


  int pointInFrustum(const gk::Vector& p) {
    for(int i=0; i < 6; i++) {
      if (volume[i].distance(p) < 0)
	return 0;
    }
    return 1;
  }


  friend std::ostream & operator <<(std::ostream &, const Camera&);
};


#endif //CAMERA2_H
