#version 430
#extension GL_ARB_texture_cube_map_array : enable

/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#ifdef VERTEX_SHADER
uniform mat4 mvpMatrix;
uniform mat4 mvMatrix;
uniform mat4 mvpMatrixInverse;
uniform mat4 depthMVP;
uniform vec4 light_direction;
uniform float fog_density;

layout (location= 0) in vec3 position;
layout (location= 1) in vec3 normal; 
layout (location= 2) in vec4 texcoord;

out vec4 vertex_normal;
out vec4 vertex_position;
out vec4 mv_position;
out vec4 vertex_texcoord;
out vec4 ShadowCoord;
out vec4 ld;
out float fogFactor; 

void main( )
{
  vertex_position = -(vec4(position,1.0) * mvMatrix);
  mv_position = (vec4(position,1.0) * mvMatrix);
  vertex_normal = vec4(normal,1.0) * mvMatrix;
  vertex_texcoord= vec4(texcoord.st,texcoord.z,texcoord.w);
  //vertex_texcoord=vec4(texcoord.xyz,texcoord.w);	
		 

  const float LOG2 = 1.442695;	
  float fogFragCoord = length(vertex_position);
  fogFactor = exp2( -fog_density * 
		    fog_density * 
		    fogFragCoord * 
		    fogFragCoord * 
		    LOG2 );

  fogFactor = clamp(fogFactor, 0.0, 1.0);

  ld = mvMatrix * light_direction;
  ShadowCoord = depthMVP * vec4(position,1.0);
  gl_Position= mvpMatrix * vec4(position, 1.0);
}
#endif

#ifdef FRAGMENT_SHADER
uniform vec3 light_position;
uniform vec4 light_color;
uniform vec3 camera_position;
uniform sampler2DArray textures;
uniform sampler2DShadow depth_texture;
uniform samplerCubeArray texture_cube_map;
uniform samplerCube cubemap;
uniform vec4 fog_color;


uniform float fogNear;
uniform float fogFar;
uniform vec3 fogColor;

in vec4 vertex_normal;
in vec4 vertex_position;
in vec4 mv_position;
in vec4 vertex_texcoord;
in vec4 ShadowCoord;
in vec4 ld;
in float fogFactor; 

out vec4 fragment_color;


void main( )
{
  vec3 texColor= texture(textures, vec3(vertex_texcoord)).rgb;
  //vec3 texColor= texture(texture_cube_map, vertex_texcoord).rgb;	
  //vec3 texColor= texture(cubemap, vec3(vertex_texcoord)).rgb;	

  //  float fFogCoord = abs(mv_position.z/mv_position.w);
  //  float fFogCoord = abs(vertex_position.z/vertex_position.w);
      
  //float visibility = texture( depth_texture, vec3(ShadowCoord.xy, (ShadowCoord.z)/ShadowCoord.w) );
  float visibility=1.0;

  vec3 l = normalize(vec3(light_position)-vec3(vertex_position));
  vec3 n = normalize(vec3(vertex_normal));
  vec3 o = normalize(camera_position-vec3(vertex_position));//vec3(0,0,0);
  vec3 lo = visibility * light_color.rgb *clamp( dot( n,l ), 0,1 ) *dot(n,o);

  vec4 color = vec4(lo + visibility * texColor /*  abs(n.z) **/,1.0);
  fragment_color = color;

  float depth = mv_position.z / mv_position.w;
  float fogFactor = smoothstep(fogNear, fogFar, depth);
  fragment_color = mix(color, vec4(fogColor, color.w), fogFactor);

  //  fragment_color = mix(vec4(lo + visibility * texColor ,1.0), fogParams.vFogColor, getFogFactor(fogParams, fFogCoord));
  //fragment_color = mix(fog_color, color, fogFactor );
  // fragment_color=vec4(1,1,1,1);
}

#endif
